@@ -123,7 +123,7 @@ update_key_down_count(struct evdev_device *device, int code, int pressed)
 	}
 
 	if (key_count > 32) {
-		log_bug_libinput(device->base.seat->libinput,
+		log_bug_libinput(evdev_libinput_context(device),
 				 "Key count for %s reached abnormal values\n",
 				 libevdev_event_code_get_name(EV_KEY, code));
 	}
@@ -370,10 +370,26 @@ evdev_filter_defuzz_touch(struct evdev_device *device, struct mt_slot *slot)
 	return false;
 }
 
+static inline void
+evdev_rotate_relative(struct evdev_device *device)
+{
+	struct evdev_dispatch *dispatch = device->dispatch;
+	struct device_coords rel = device->rel;
+
+	if (!device->base.config.rotation)
+		return;
+
+	/* loss of precision for non-90 degrees, but we only support 90 deg
+	 * right now anyway */
+	matrix_mult_vec(&dispatch->rotation.matrix, &rel.x, &rel.y);
+
+	device->rel = rel;
+}
+
 static void
 evdev_flush_pending_event(struct evdev_device *device, uint64_t time)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	int slot_idx;
 	int seat_slot;
 	struct libinput_device *base = &device->base;
@@ -394,6 +410,8 @@ evdev_flush_pending_event(struct evdev_device *device, uint64_t time)
 		if (!(device->seat_caps & EVDEV_DEVICE_POINTER))
 			break;
 
+		evdev_rotate_relative(device);
+
 		normalize_delta(device, &device->rel, &unaccel);
 		raw.x = device->rel.x;
 		raw.y = device->rel.y;
@@ -655,7 +673,7 @@ evdev_process_touch(struct evdev_device *device,
 	switch (e->code) {
 	case ABS_MT_SLOT:
 		if ((size_t)e->value >= device->mt.slots_len) {
-			log_bug_libinput(device->base.seat->libinput,
+			log_bug_libinput(evdev_libinput_context(device),
 					 "%s exceeds slots (%d of %zd)\n",
 					 device->devname,
 					 e->value,
@@ -736,11 +754,9 @@ evdev_reject_relative(struct evdev_device *device,
 		      const struct input_event *e,
 		      uint64_t time)
 {
-	struct libinput *libinput = device->base.seat->libinput;
-
 	if ((e->code == REL_X || e->code == REL_Y) &&
 	    (device->seat_caps & EVDEV_DEVICE_POINTER) == 0) {
-		log_bug_libinput_ratelimit(libinput,
+		log_bug_libinput_ratelimit(evdev_libinput_context(device),
 					   &device->nonpointer_rel_limit,
 					   "REL_X/Y from device '%s', but this device is not a pointer\n",
 					   device->devname);
@@ -919,7 +935,7 @@ fallback_process(struct evdev_dispatch *dispatch,
 static void
 release_pressed_keys(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	uint64_t time;
 	int code;
 
@@ -1242,7 +1258,8 @@ static int
 evdev_init_button_scroll(struct evdev_device *device,
 			 void (*change_scroll_method)(struct evdev_device *))
 {
-	libinput_timer_init(&device->scroll.timer, device->base.seat->libinput,
+	libinput_timer_init(&device->scroll.timer,
+			    evdev_libinput_context(device),
 			    evdev_button_scroll_timeout, device);
 	device->scroll.config.get_methods = evdev_scroll_get_methods;
 	device->scroll.config.set_method = evdev_scroll_set_method;
@@ -1330,6 +1347,55 @@ evdev_init_natural_scroll(struct evdev_device *device)
 	device->base.config.natural_scroll = &device->scroll.config_natural;
 }
 
+static int
+evdev_rotation_config_is_available(struct libinput_device *device)
+{
+	/* This function only gets called when we support rotation */
+	return 1;
+}
+
+static enum libinput_config_status
+evdev_rotation_config_set_angle(struct libinput_device *device,
+				unsigned int degrees_cw)
+{
+	struct evdev_dispatch *dispatch = ((struct evdev_device*)device)->dispatch;
+
+	dispatch->rotation.angle = degrees_cw;
+	matrix_init_rotate(&dispatch->rotation.matrix, degrees_cw);
+
+	return LIBINPUT_CONFIG_STATUS_SUCCESS;
+}
+
+static unsigned int
+evdev_rotation_config_get_angle(struct libinput_device *device)
+{
+	struct evdev_dispatch *dispatch = ((struct evdev_device*)device)->dispatch;
+
+	return dispatch->rotation.angle;
+}
+
+static unsigned int
+evdev_rotation_config_get_default_angle(struct libinput_device *device)
+{
+	return 0;
+}
+
+static void
+evdev_init_rotation(struct evdev_device *device,
+		    struct evdev_dispatch *dispatch)
+{
+	if ((device->model_flags & EVDEV_MODEL_TRACKBALL) == 0)
+		return;
+
+	dispatch->rotation.config.is_available = evdev_rotation_config_is_available;
+	dispatch->rotation.config.set_angle = evdev_rotation_config_set_angle;
+	dispatch->rotation.config.get_angle = evdev_rotation_config_get_angle;
+	dispatch->rotation.config.get_default_angle = evdev_rotation_config_get_default_angle;
+	dispatch->rotation.is_enabled = false;
+	matrix_init_identity(&dispatch->rotation.matrix);
+	device->base.config.rotation = &dispatch->rotation.config;
+}
+
 static struct evdev_dispatch *
 fallback_dispatch_create(struct libinput_device *device)
 {
@@ -1360,6 +1426,7 @@ fallback_dispatch_create(struct libinput_device *device)
 
 	evdev_init_calibration(evdev_device, dispatch);
 	evdev_init_sendevents(evdev_device, dispatch);
+	evdev_init_rotation(evdev_device, dispatch);
 
 	/* BTN_MIDDLE is set on mice even when it's not present. So
 	 * we can only use the absence of BTN_MIDDLE to mean something, i.e.
@@ -1390,10 +1457,10 @@ evdev_process_event(struct evdev_device *device, struct input_event *e)
 
 #if 0
 	if (libevdev_event_is_code(e, EV_SYN, SYN_REPORT))
-		log_debug(device->base.seat->libinput,
+		log_debug(evdev_libinput_context(device),
 			  "-------------- EV_SYN ------------\n");
 	else
-		log_debug(device->base.seat->libinput,
+		log_debug(evdev_libinput_context(device),
 			  "%-7s %-16s %-20s %4d\n",
 			  evdev_device_get_sysname(device),
 			  libevdev_event_type_get_name(e->type),
@@ -1443,7 +1510,7 @@ static void
 evdev_device_dispatch(void *data)
 {
 	struct evdev_device *device = data;
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	struct input_event ev;
 	int rc;
 
@@ -1627,7 +1694,6 @@ evdev_need_mtdev(struct evdev_device *device)
 static inline int
 evdev_read_wheel_click_prop(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
 	const char *prop;
 	int angle = DEFAULT_WHEEL_CLICK_ANGLE;
 
@@ -1636,7 +1702,7 @@ evdev_read_wheel_click_prop(struct evdev_device *device)
 	if (prop) {
 		angle = parse_mouse_wheel_click_angle_property(prop);
 		if (!angle) {
-			log_error(libinput,
+			log_error(evdev_libinput_context(device),
 				  "Mouse wheel click angle '%s' is present but invalid,"
 				  "using %d degrees instead\n",
 				  device->devname,
@@ -1651,7 +1717,7 @@ evdev_read_wheel_click_prop(struct evdev_device *device)
 static inline int
 evdev_get_trackpoint_dpi(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	const char *trackpoint_accel;
 	double accel = DEFAULT_TRACKPOINT_ACCEL;
 
@@ -1679,7 +1745,7 @@ evdev_get_trackpoint_dpi(struct evdev_device *device)
 static inline int
 evdev_read_dpi_prop(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	const char *mouse_dpi;
 	int dpi = DEFAULT_MOUSE_DPI;
 
@@ -1719,24 +1785,30 @@ evdev_read_model_flags(struct evdev_device *device)
 		const char *property;
 		enum evdev_device_model model;
 	} model_map[] = {
-		{ "LIBINPUT_MODEL_LENOVO_X230", EVDEV_MODEL_LENOVO_X230 },
-		{ "LIBINPUT_MODEL_LENOVO_X220_TOUCHPAD_FW81", EVDEV_MODEL_LENOVO_X220_TOUCHPAD_FW81 },
-		{ "LIBINPUT_MODEL_CHROMEBOOK", EVDEV_MODEL_CHROMEBOOK },
-		{ "LIBINPUT_MODEL_SYSTEM76_BONOBO", EVDEV_MODEL_SYSTEM76_BONOBO },
-		{ "LIBINPUT_MODEL_SYSTEM76_GALAGO", EVDEV_MODEL_SYSTEM76_GALAGO },
-		{ "LIBINPUT_MODEL_SYSTEM76_KUDU", EVDEV_MODEL_SYSTEM76_KUDU },
-		{ "LIBINPUT_MODEL_CLEVO_W740SU", EVDEV_MODEL_CLEVO_W740SU },
-		{ "LIBINPUT_MODEL_APPLE_TOUCHPAD", EVDEV_MODEL_APPLE_TOUCHPAD },
-		{ "LIBINPUT_MODEL_WACOM_TOUCHPAD", EVDEV_MODEL_WACOM_TOUCHPAD },
-		{ "LIBINPUT_MODEL_ALPS_TOUCHPAD", EVDEV_MODEL_ALPS_TOUCHPAD },
-		{ "LIBINPUT_MODEL_SYNAPTICS_SERIAL_TOUCHPAD", EVDEV_MODEL_SYNAPTICS_SERIAL_TOUCHPAD },
-		{ "LIBINPUT_MODEL_JUMPING_SEMI_MT", EVDEV_MODEL_JUMPING_SEMI_MT },
-		{ "LIBINPUT_MODEL_ELANTECH_TOUCHPAD", EVDEV_MODEL_ELANTECH_TOUCHPAD },
-		{ "LIBINPUT_MODEL_APPLE_INTERNAL_KEYBOARD", EVDEV_MODEL_APPLE_INTERNAL_KEYBOARD },
-		{ "LIBINPUT_MODEL_CYBORG_RAT", EVDEV_MODEL_CYBORG_RAT },
-		{ "LIBINPUT_MODEL_CYAPA", EVDEV_MODEL_CYAPA },
-		{ "LIBINPUT_MODEL_LENOVO_T450_TOUCHPAD", EVDEV_MODEL_LENOVO_T450_TOUCHPAD },
+#define MODEL(name) { "LIBINPUT_MODEL_" #name, EVDEV_MODEL_##name }
+		MODEL(LENOVO_X230),
+		MODEL(LENOVO_X230),
+		MODEL(LENOVO_X220_TOUCHPAD_FW81),
+		MODEL(CHROMEBOOK),
+		MODEL(SYSTEM76_BONOBO),
+		MODEL(SYSTEM76_GALAGO),
+		MODEL(SYSTEM76_KUDU),
+		MODEL(CLEVO_W740SU),
+		MODEL(APPLE_TOUCHPAD),
+		MODEL(WACOM_TOUCHPAD),
+		MODEL(ALPS_TOUCHPAD),
+		MODEL(SYNAPTICS_SERIAL_TOUCHPAD),
+		MODEL(JUMPING_SEMI_MT),
+		MODEL(ELANTECH_TOUCHPAD),
+		MODEL(APPLE_INTERNAL_KEYBOARD),
+		MODEL(CYBORG_RAT),
+		MODEL(CYAPA),
+		MODEL(LENOVO_T450_TOUCHPAD),
+		MODEL(DELL_TOUCHPAD),
+		MODEL(TRACKBALL),
+		MODEL(APPLE_MAGICMOUSE),
 		{ NULL, EVDEV_MODEL_DEFAULT },
+#undef MODEL
 	};
 	const struct model_map *m = model_map;
 	uint32_t model_flags = 0;
@@ -1746,7 +1818,7 @@ evdev_read_model_flags(struct evdev_device *device)
 		val = udev_device_get_property_value(device->udev_device,
 						     m->property);
 		if (val && !streq(val, "0")) {
-			log_debug(device->base.seat->libinput,
+			log_debug(evdev_libinput_context(device),
 				  "%s: tagged as %s\n",
 				  evdev_device_get_sysname(device),
 				  m->property);
@@ -1798,7 +1870,6 @@ evdev_fix_abs_resolution(struct evdev_device *device,
 			 unsigned int xcode,
 			 unsigned int ycode)
 {
-	struct libinput *libinput = device->base.seat->libinput;
 	struct libevdev *evdev = device->evdev;
 	const struct input_absinfo *absx, *absy;
 	size_t widthmm = 0, heightmm = 0;
@@ -1807,7 +1878,7 @@ evdev_fix_abs_resolution(struct evdev_device *device,
 
 	if (!(xcode == ABS_X && ycode == ABS_Y)  &&
 	    !(xcode == ABS_MT_POSITION_X && ycode == ABS_MT_POSITION_Y)) {
-		log_bug_libinput(libinput,
+		log_bug_libinput(evdev_libinput_context(device),
 				 "Invalid x/y code combination %d/%d\n",
 				 xcode, ycode);
 		return 0;
@@ -1914,7 +1985,7 @@ evdev_check_min_max(struct evdev_device *device, unsigned int code)
 		 */
 		if (absinfo->minimum == 0 &&
 		    code >= ABS_MISC && code < ABS_MT_SLOT) {
-			log_info(device->base.seat->libinput,
+			log_info(evdev_libinput_context(device),
 				 "Disabling EV_ABS %#x on device '%s' (min == max == 0)\n",
 				 code,
 				 device->devname);
@@ -1922,7 +1993,7 @@ evdev_check_min_max(struct evdev_device *device, unsigned int code)
 						    EV_ABS,
 						    code);
 		} else {
-			log_bug_kernel(device->base.seat->libinput,
+			log_bug_kernel(evdev_libinput_context(device),
 				       "Device '%s' has min == max on %s\n",
 				       device->devname,
 				       libevdev_event_code_get_name(EV_ABS, code));
@@ -1936,7 +2007,7 @@ evdev_check_min_max(struct evdev_device *device, unsigned int code)
 static int
 evdev_reject_device(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	struct libevdev *evdev = device->evdev;
 	unsigned int code;
 	const struct input_absinfo *absx, *absy;
@@ -2068,7 +2139,7 @@ evdev_configure_mt_device(struct evdev_device *device)
 static int
 evdev_configure_device(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	struct libevdev *evdev = device->evdev;
 	const char *devnode = udev_device_get_devnode(device->udev_device);
 	enum evdev_device_udev_tags udev_tags;
@@ -2173,7 +2244,6 @@ evdev_configure_device(struct evdev_device *device)
 			 "input device '%s', %s is a touchpad\n",
 			 device->devname, devnode);
 
-		evdev_tag_touchpad(device, device->udev_device);
 		return device->dispatch == NULL ? -1 : 0;
 	}
 
@@ -2290,7 +2360,7 @@ static int
 evdev_set_device_group(struct evdev_device *device,
 		       struct udev_device *udev_device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	struct libinput_device_group *group = NULL;
 	const char *udev_group;
 
@@ -2354,6 +2424,12 @@ evdev_pre_configure_model_quirks(struct evdev_device *device)
 		libevdev_disable_event_code(device->evdev, EV_KEY, 0x119);
 		libevdev_disable_event_code(device->evdev, EV_KEY, 0x11a);
 	}
+	/* The Apple MagicMouse has a touchpad built-in but the kernel still
+	 * emulates a full 2/3 button mouse for us. Ignore anything from the
+	 * ABS interface
+	 */
+	if (device->model_flags & EVDEV_MODEL_APPLE_MAGICMOUSE)
+		libevdev_disable_event_type(device->evdev, EV_ABS);
 }
 
 struct evdev_device *
@@ -2595,7 +2671,7 @@ evdev_device_has_capability(struct evdev_device *device,
 }
 
 int
-evdev_device_get_size(struct evdev_device *device,
+evdev_device_get_size(const struct evdev_device *device,
 		      double *width,
 		      double *height)
 {
@@ -2792,6 +2868,8 @@ evdev_notify_resumed_device(struct evdev_device *device)
 int
 evdev_device_suspend(struct evdev_device *device)
 {
+	struct libinput *libinput = evdev_libinput_context(device);
+
 	evdev_notify_suspended_device(device);
 
 	if (device->dispatch->interface->suspend)
@@ -2799,8 +2877,7 @@ evdev_device_suspend(struct evdev_device *device)
 						     device);
 
 	if (device->source) {
-		libinput_remove_source(device->base.seat->libinput,
-				       device->source);
+		libinput_remove_source(libinput, device->source);
 		device->source = NULL;
 	}
 
@@ -2810,7 +2887,7 @@ evdev_device_suspend(struct evdev_device *device)
 	}
 
 	if (device->fd != -1) {
-		close_restricted(device->base.seat->libinput, device->fd);
+		close_restricted(libinput, device->fd);
 		device->fd = -1;
 	}
 
@@ -2820,7 +2897,7 @@ evdev_device_suspend(struct evdev_device *device)
 int
 evdev_device_resume(struct evdev_device *device)
 {
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	int fd;
 	const char *devnode;
 	struct input_event ev;
@@ -2936,7 +3013,7 @@ evdev_tablet_has_left_handed(struct evdev_device *device)
 {
 	bool has_left_handed = false;
 #if HAVE_LIBWACOM
-	struct libinput *libinput = device->base.seat->libinput;
+	struct libinput *libinput = evdev_libinput_context(device);
 	WacomDeviceDatabase *db;
 	WacomDevice *d = NULL;
 	WacomError *error;
